package explorviz.live_trace_processing.filter.reduction;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.TreeMap;

import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.SinglePipeConnector;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractBeforeEventRecord;
import explorviz.live_trace_processing.record.event.AbstractBeforeOperationEventRecord;
import explorviz.live_trace_processing.record.event.AbstractEventRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.RuntimeStatisticInformation;
import explorviz.live_trace_processing.record.trace.Trace;
import explorviz.live_trace_processing.record.trace.TraceComperator;

public class TracesSummarizationFilter extends AbstractFilter {
	// private static final String INVALID_TRACE = "invalid trace...";

	private final long maxCollectionDuration;

	private final Map<Trace, TracesSummarizationBuffer> trace2buffer = new TreeMap<Trace, TracesSummarizationBuffer>(
			new TraceComperator());

	private final SinglePipeConnector<IRecord> pipeConnector;

	public TracesSummarizationFilter(final SinglePipeConnector<IRecord> pipeConnector,
			final long maxCollectionDuration, final Queue<IRecord> sinkReceiver) {
		super(sinkReceiver, "Reduced traces/sec", 1000);
		this.pipeConnector = pipeConnector;

		this.maxCollectionDuration = maxCollectionDuration;
	}

	@Override
	public void run() {
		pipeConnector.process(this);
	}

	@Override
	public void processRecord(final IRecord record) {
		if (record instanceof Trace) {
			final Trace trace = (Trace) record;

			if (trace.isValid() && !trace.containsRemoteRecord()) {
				insertIntoBuffer(trace);
			} else {
				// trace with remote records or invalid trace cant be reduced
				// System.out.println(INVALID_TRACE);
				// System.out.println("invalid trace... trace:" +
				// trace.toString());
				makeTraceElementsAccumulator(trace);
				deliver(trace);
			}
		} else if (record instanceof TimedPeriodRecord) {
			processTimeoutQueue(TimeProvider.getCurrentTimestamp());
			deliver(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
			deliver(record);
		} else {
			deliver(record);
		}
	}

	private void insertIntoBuffer(final Trace trace) {
		TracesSummarizationBuffer traceAggregationBuffer = trace2buffer.get(trace);
		if (traceAggregationBuffer == null) {
			makeTraceElementsAccumulator(trace);

			traceAggregationBuffer = new TracesSummarizationBuffer(
					TimeProvider.getCurrentTimestamp());
			trace2buffer.put(trace, traceAggregationBuffer);
		}
		traceAggregationBuffer.insertTrace(trace);
	}

	private void makeTraceElementsAccumulator(final Trace trace) {
		final List<AbstractEventRecord> traceEvents = trace.getTraceEvents();

		for (final AbstractEventRecord event : traceEvents) {
			if (event instanceof AbstractBeforeEventRecord) {
				final AbstractBeforeEventRecord abstractBeforeEventRecord = (AbstractBeforeEventRecord) event;

				final List<RuntimeStatisticInformation> runtimeStatisticInformationList = abstractBeforeEventRecord
						.getRuntimeStatisticInformationList();
				if (runtimeStatisticInformationList.size() == 1) {
					final RuntimeStatisticInformation runtimeStatisticInformation = runtimeStatisticInformationList
							.get(0);
					if (!runtimeStatisticInformation.isInitialized()) {
						int objectId = 0;

						if (abstractBeforeEventRecord instanceof AbstractBeforeOperationEventRecord) {
							final AbstractBeforeOperationEventRecord abstractBeforeOperationEventRecord = (AbstractBeforeOperationEventRecord) abstractBeforeEventRecord;
							objectId = abstractBeforeOperationEventRecord.getObjectId();
						}

						runtimeStatisticInformation.makeAccumulator(objectId);
					}
				}
			}
		}
	}

	private void processTimeoutQueue(final long timestamp) {
		final long bufferTimeout = timestamp - maxCollectionDuration;
		final Iterator<Entry<Trace, TracesSummarizationBuffer>> iter = trace2buffer.entrySet()
				.iterator();

		while (iter.hasNext()) {
			final Entry<Trace, TracesSummarizationBuffer> traceBufferEntry = iter.next();

			if (traceBufferEntry.getValue().getBufferCreatedTimestamp() <= bufferTimeout) {
				final Trace aggregatedTrace = traceBufferEntry.getValue().getAggregatedTrace();
				deliver(aggregatedTrace);
				iter.remove();
			}
		}
	}

	private void terminate() {
		for (final TracesSummarizationBuffer traceBuffer : trace2buffer.values()) {
			deliver(traceBuffer.getAggregatedTrace());
		}
		trace2buffer.clear();
	}
}
